/*
 * kgraft_patch_bsc980883
 *
 * Fix for CVE-2016-4565, bsc#980883, ib_uverbs module
 *
 *  Upstream commit:
 *  e6bd18f57aad ("IB/security: Restrict use of the write() interface")
 *
 *  SLE12 commit:
 *  47463e845fb78e44ff63447c8c575aa4737bf7ed
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/fs.h>
#include <rdma/ib_verbs.h>
#include <rdma/ib_user_verbs.h>
#include <asm/uaccess.h>

#include "local.h"

#define KGR_PATCHED_MODULE	"ib_uverbs"

static ssize_t (*(*kgr_uverbs_cmd_table)[])(struct ib_uverbs_file *file, const char __user *buf, int in_len, int out_len);
static int (*(*kgr_uverbs_ex_cmd_table)[])(struct ib_uverbs_file *file, struct ib_udata *ucore, struct ib_udata *uhw);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "uverbs_cmd_table", (void *)&kgr_uverbs_cmd_table },
	{ "uverbs_ex_cmd_table", (void *)&kgr_uverbs_ex_cmd_table },
};

/* patched function */
ssize_t kgr_ib_uverbs_write(struct file *filp, const char __user *buf,
	size_t count, loff_t *pos)
{
	struct ib_uverbs_file *file = filp->private_data;
	struct ib_uverbs_cmd_hdr hdr;
	__u32 flags;

	/*
	 * Fix CVE-2016-4546
	 *  +next 2 lines
	 */
	if (WARN_ON_ONCE(!kgr_ib_safe_file_access(filp)))
		return -EACCES;

	if (count < sizeof hdr)
		return -EINVAL;

	if (copy_from_user(&hdr, buf, sizeof hdr))
		return -EFAULT;

	flags = (hdr.command &
		 IB_USER_VERBS_CMD_FLAGS_MASK) >> IB_USER_VERBS_CMD_FLAGS_SHIFT;

	if (!flags) {
		__u32 command;

		if (hdr.command & ~(__u32)(IB_USER_VERBS_CMD_FLAGS_MASK |
					   IB_USER_VERBS_CMD_COMMAND_MASK))
			return -EINVAL;

		command = hdr.command & IB_USER_VERBS_CMD_COMMAND_MASK;

		/*
		 * kgr: hardcode the array size. It is the same across all maintenance
		 * updates.
		 * -if (command >= ARRAY_SIZE(uverbs_cmd_table) ||
		 */
		if (command >= 41 ||
		    !(*kgr_uverbs_cmd_table)[command])
			return -EINVAL;

		if (!file->ucontext &&
		    command != IB_USER_VERBS_CMD_GET_CONTEXT)
			return -EINVAL;

		if (!(file->device->ib_dev->uverbs_cmd_mask & (1ull << command)))
			return -ENOSYS;

		if (hdr.in_words * 4 != count)
			return -EINVAL;

		return (*kgr_uverbs_cmd_table)[command](file,
						 buf + sizeof(hdr),
						 hdr.in_words * 4,
						 hdr.out_words * 4);

	} else if (flags == IB_USER_VERBS_CMD_FLAG_EXTENDED) {
		__u32 command;

		struct ib_uverbs_ex_cmd_hdr ex_hdr;
		struct ib_udata ucore;
		struct ib_udata uhw;
		int err;
		size_t written_count = count;

		if (hdr.command & ~(__u32)(IB_USER_VERBS_CMD_FLAGS_MASK |
					   IB_USER_VERBS_CMD_COMMAND_MASK))
			return -EINVAL;

		command = hdr.command & IB_USER_VERBS_CMD_COMMAND_MASK;

		/*
		 * kgr: hardcode the array size. It is the same across all maintenance
		 * updates.
		 * -if (command >= ARRAY_SIZE(uverbs_ex_cmd_table) ||
		 */
		if (command >= 52 ||
		    !(*kgr_uverbs_ex_cmd_table)[command])
			return -ENOSYS;

		if (!file->ucontext)
			return -EINVAL;

		if (!(file->device->ib_dev->uverbs_ex_cmd_mask & (1ull << command)))
			return -ENOSYS;

		if (count < (sizeof(hdr) + sizeof(ex_hdr)))
			return -EINVAL;

		if (copy_from_user(&ex_hdr, buf + sizeof(hdr), sizeof(ex_hdr)))
			return -EFAULT;

		count -= sizeof(hdr) + sizeof(ex_hdr);
		buf += sizeof(hdr) + sizeof(ex_hdr);

		if ((hdr.in_words + ex_hdr.provider_in_words) * 8 != count)
			return -EINVAL;

		if (ex_hdr.cmd_hdr_reserved)
			return -EINVAL;

		if (ex_hdr.response) {
			if (!hdr.out_words && !ex_hdr.provider_out_words)
				return -EINVAL;

			if (!access_ok(VERIFY_WRITE,
				       (void __user *) (unsigned long) ex_hdr.response,
				       (hdr.out_words + ex_hdr.provider_out_words) * 8))
				return -EFAULT;
		} else {
			if (hdr.out_words || ex_hdr.provider_out_words)
				return -EINVAL;
		}

		INIT_UDATA(&ucore,
			   (hdr.in_words) ? buf : 0,
			   (unsigned long)ex_hdr.response,
			   hdr.in_words * 8,
			   hdr.out_words * 8);

		INIT_UDATA(&uhw,
			   (ex_hdr.provider_in_words) ? buf + ucore.inlen : 0,
			   (ex_hdr.provider_out_words) ? (unsigned long)ex_hdr.response + ucore.outlen : 0,
			   ex_hdr.provider_in_words * 8,
			   ex_hdr.provider_out_words * 8);

		err = (*kgr_uverbs_ex_cmd_table)[command](file,
						   &ucore,
						   &uhw);

		if (err)
			return err;

		return written_count;
	}

	return -ENOSYS;
}

static int kgr_patch_bsc980883_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc980883_module_notify(struct notifier_block *nb,
        unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc980883_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc980883_module_nb = {
	.notifier_call = kgr_patch_bsc980883_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc980883_ibuverbs_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc980883_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc980883_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc980883_ibuverbs_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc980883_module_nb);
}
